unit JpgImg;

{$define TWO_FILE_COMMANDLINE}
{$define BMP_SUPPORTED}

{$I jconfig.inc}

interface

uses
  SysUtils,
  Classes,
  ExtCtrls,
  DsgnIntf, Dialogs, {for property editor}
  jmorecfg,
  jpeglib,
  jerror,
  jdeferr,
  jdapimin, jdapistd, jdatasrc, wrbmp, 
  cdjpeg;		{ Common decls for cjpeg/djpeg applications }



type
  EJPEGException = class(Exception);
  TNumOfColors = (_16, _256, TrueColor);
  TDCTMethod = (Integ, FastInteg, FloatingPoint);
  TDither = (None, FS, Ordered);
  TDelta = 1..100;

  IMAGE_FORMATS = (
	FMT_BMP,		{ BMP format (Windows flavor) }
	FMT_GIF,		{ GIF format }
	FMT_OS2,		{ BMP format (OS/2 flavor) }
	FMT_PPM,		{ PPM/PGM (PBMPLUS formats) }
	FMT_RLE,		{ RLE format }
	FMT_TARGA,		{ Targa format }
	FMT_TIFF);		{ TIFF format }

  TJPEGNameProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

  TJPEGImage = class(TImage)
  private
    FNumOfColors: TNumOfColors;
    FDCTMethod: TDCTMethod;
    FDither: TDither;
    FFastProcessing, FGrayScale, FNoSmooth, FOnePass: Boolean;
    FFileName: ShortString;
    FProgressStep: TDelta;
    FOnProgress: TNotifyEvent;
    OutStream: TMemoryStream;

    {Procs for properties}
    procedure SetNumOfColors(Value: TNumOfColors);
    procedure SetDCTMethod(Value: TDCTMethod);
    procedure SetDither(Value: TDither);
    procedure SetBoolValue(Index: Integer; Value: Boolean);
    procedure SetFileName(Value: ShortString);
  protected
    procedure parse_switches (cinfo : j_decompress_ptr;
                          last_file_arg_seen : int;
                          for_real : boolean);

  public
    {Constructor/destructor}
    constructor Create(AOwner: TComponent); override;

  published
    property NumOfColors: TNumOfColors read FNumOfColors write SetNumOfColors;
    property DCTMethod: TDCTMethod read FDCTMethod write SetDCTMethod;
    property Dither: TDither read FDither write SetDither;
    property FastProcessing: Boolean index 0 read FFastProcessing write SetBoolValue;
    property GrayScale: Boolean index 1 read FGrayScale write SetBoolValue;
    property NoSmooth: Boolean index 2 read FNoSmooth write SetBoolValue;
    property OnePass: Boolean index 3 read FOnePass write SetBoolValue;
    property FileName: ShortString read FFileName write SetFileName;
    property ProgressStep: TDelta read FProgressStep write FProgressStep;
    property OnProgress: TNotifyEvent read FOnProgress write FOnProgress;
  end;

const
  DEFAULT_FMT = FMT_PPM;

var
  requested_fmt : IMAGE_FORMATS;

procedure Register;

{Library procedures and functions}
function jpeg_getc (cinfo : j_decompress_ptr) : char;
function COM_handler (cinfo : j_decompress_ptr) : boolean;

implementation

procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(ShortString), TJPEGImage, 'FileName', TJPEGNameProperty);
  RegisterComponents('Samples', [TJPEGImage]);
end;

function TJPEGNameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paMultiSelect, paAutoUpdate];
end;

procedure TJPEGNameProperty.Edit;
begin
  with TOpenDialog.Create(nil) do begin
    DefaultExt := 'jpg';
    Filter := 'JPeg Files (*.jpg)|*.jpg|All Files (*.*)|*.*';
    Options := [ofHideReadOnly,ofPathMustExist,ofFileMustExist];
    FileName := GetStrValue;
    Title := 'Select JPeg File';
    if Execute then SetStrValue(FileName);
    Free;
  end;
end;

constructor TJPEGImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FileName := '';
  NumOfColors := _256;
  DCTMethod := Integ;
  Dither := FS;
  GrayScale := False;
  NoSmooth := False;
  OnePass := False;
  FProgressStep := 10;
end;

procedure TJPegImage.parse_switches (cinfo : j_decompress_ptr;
		                    last_file_arg_seen : int;
                                    for_real : boolean);
const
  printed_version : boolean = FALSE;
begin
  { Set up default JPEG parameters. }
  requested_fmt := DEFAULT_FMT;	{ set default output file format }
  cinfo^.err^.trace_level := 0;
  { BMP output format. }
  requested_fmt := FMT_BMP;
  { Num of colors }
  if FNumOfColors <> TrueColor then
  begin
    case FNumOfColors of
          _16: cinfo^.desired_number_of_colors := 16;
         _256: cinfo^.desired_number_of_colors := 256;
    end;
    cinfo^.quantize_colors := TRUE;
  end;
  { DCT Method }
  case DCTMethod of
            Integ: cinfo^.dct_method := JDCT_ISLOW;
        FastInteg: cinfo^.dct_method := JDCT_IFAST;
    FloatingPoint: cinfo^.dct_method := JDCT_FLOAT;
  end;
  { Dither }
  case Dither of
       None: cinfo^.dither_mode := JDITHER_NONE;
         FS: cinfo^.dither_mode := JDITHER_FS;
    Ordered: cinfo^.dither_mode := JDITHER_ORDERED;
  end;
  { Fast Processing }
  if FFastProcessing then
  begin
    { Select recommended processing options for quick-and-dirty output. }
    cinfo^.two_pass_quantize := FALSE;
    cinfo^.dither_mode := JDITHER_ORDERED;
    if (not cinfo^.quantize_colors) then { don't override an earlier -colors }
      cinfo^.desired_number_of_colors := 216;
    cinfo^.dct_method := JDCT_FASTEST;
    cinfo^.do_fancy_upsampling := FALSE;
  end;
  { Grayscale }
  if FGrayScale then
    { Force monochrome output. }
    cinfo^.out_color_space := JCS_GRAYSCALE;
  if FNoSmooth then
    { Suppress fancy upsampling }
    cinfo^.do_fancy_upsampling := FALSE;
  if FOnePass then
    { Use fast one-pass quantization. }
    cinfo^.two_pass_quantize := FALSE;
end;

function jpeg_getc (cinfo : j_decompress_ptr) : char;
{ Read next byte }
var
  datasrc : jpeg_source_mgr_ptr;
begin
  datasrc := cinfo^.src;

  if (datasrc^.bytes_in_buffer = 0) then
  begin
    if (not datasrc^.fill_input_buffer (cinfo)) then
      ERREXIT(j_common_ptr(cinfo), JERR_CANT_SUSPEND);
  end;
  Dec(datasrc^.bytes_in_buffer);
  jpeg_getc := char(GETJOCTET(datasrc^.next_input_byte^));
  Inc(datasrc^.next_input_byte);
end;

{METHODDEF}
function COM_handler (cinfo : j_decompress_ptr) : boolean; {far;}
const
  LF = #10;
  CR = #13;
var
  length : INT32;
  ch : char;
  lastch : char;
begin
  lastch := #0;

  length := byte(jpeg_getc(cinfo)) shl 8;
  Inc(length, byte(jpeg_getc(cinfo)));
  Dec(length, 2);			{ discount the length word itself }

  while (length > 0) do
  begin
    Dec(length);
    ch := jpeg_getc(cinfo);
  end;
  COM_handler := TRUE;
end;

procedure TJPEGImage.SetNumOfColors(Value: TNumOfColors);
begin
  FNumOfColors := Value;
  SetFileName(FFileName);                                  {trick to refresh}
end;

procedure TJPEGImage.SetDCTMethod(Value: TDCTMethod);
begin
  FDCTMethod := Value;
  SetNumOfColors(FNumOfColors);                            {trick to refresh}
end;

procedure TJPEGImage.SetDither(Value: TDither);
begin
  FDither := Value;
  SetNumOfColors(FNumOfColors);                            {trick to refresh}
end;

procedure TJPEGImage.SetBoolValue(Index: Integer; Value: Boolean);
begin
  case Index of
    0: FFastProcessing := Value;
    1: FGrayScale := Value;
    2: FNoSmooth := Value;
    3: FOnePass := Value;
  end;
  SetNumOfColors(FNumOfColors);                            {trick to refresh}
end;

procedure TJpegImage.SetFileName(Value: ShortString);
var
  cinfo : jpeg_decompress_struct;
  jerr : jpeg_error_mgr;
  dest_mgr : djpeg_dest_ptr;
  input_file : FILE;
  num_scanlines : JDIMENSION;
  OldProgress, NewProgress: Integer;
begin
  FFileName := Value;
  if (Value = '') or (csReading in ComponentState) then
    Exit;
  dest_mgr := NIL;
  { Initialize the JPEG decompression object with default error handling. }
  cinfo.err := jpeg_std_error(jerr);
  jpeg_create_decompress(@cinfo);
  { Add some application-specific error messages (from cderror.h) }
  {jerr.addon_message_table := cdjpeg_message_table;}
  jerr.first_addon_message := JMSG_FIRSTADDONCODE;
  jerr.last_addon_message := JMSG_LASTADDONCODE;
  { Insert custom COM marker processor. }
  jpeg_set_marker_processor(@cinfo, JPEG_COM, COM_handler);

  { Now safe to enable signal catcher. }
{$ifdef NEED_SIGNAL_CATCHER}
  enable_signal_catcher(j_common_ptr (@cinfo));
{$endif}

  { Scan command line to find file names. }
  { It is convenient to use just one switch-parsing routine, but the switch
    values read here are ignored; we will rescan the switches after opening
    the input file.
    (Exception: tracing level set here controls verbosity for COM markers
    found during jpeg_read_header...) }

  parse_switches(@cinfo, 0, FALSE);

  { Open the input file. }
  System.assign(input_file, FFileName);
  {$I-}
  Reset(input_file, 1);
  {$I+}
  if (IOresult <> 0) then
    raise EFOpenError.Create('Cannot open file ' + FFileName);
  { Open the output stream. }
  try
    OutStream := TMemoryStream.Create;

    { Specify data source for decompression }
    jpeg_stdio_src(@cinfo, input_file);

    { Read file header, set default decompression parameters }
    {void} jpeg_read_header(@cinfo, TRUE);

    { Adjust default decompression parameters by re-parsing the options }
    parse_switches(@cinfo, 0, TRUE);

    { Initialize the output module now to let it override any crucial
      option settings (for instance, GIF wants to force color quantization). }

    case (requested_fmt) of
  {$ifdef BMP_SUPPORTED}
    FMT_BMP:
      dest_mgr := jinit_write_bmp(@cinfo, FALSE);
  {$endif}
    else
      ERREXIT(j_common_ptr(@cinfo), JERR_UNSUPPORTED_FORMAT);
    end;
    {dest_mgr^.output_file := @output_file;}
     dest_mgr^.output_file := @OutStream;

    { Start decompressor }
    {void} jpeg_start_decompress(@cinfo);

    { Write output file header }
    dest_mgr^.start_output (@cinfo, dest_mgr);

    OldProgress := 0;
    { Process data }
    while (cinfo.output_scanline < cinfo.output_height) do
    begin
      num_scanlines := jpeg_read_scanlines(@cinfo, dest_mgr^.buffer,
                                          dest_mgr^.buffer_height);
      dest_mgr^.put_pixel_rows (@cinfo, dest_mgr, num_scanlines);
      NewProgress := (cinfo.output_scanline * 100 div cinfo.output_height);
      if (NewProgress <> OldProgress) and (NewProgress mod FProgressStep = 0) then
      begin
        OldProgress := NewProgress;
        if Assigned(FOnProgress) then
          FOnProgress(Self);
      end;
    end;

    { Finish decompression and release memory.
      I must do it in this order because output module has allocated memory
      of lifespan JPOOL_IMAGE; it needs to finish before releasing memory. }

    dest_mgr^.finish_output (@cinfo, dest_mgr);
    {void} jpeg_finish_decompress(@cinfo);
    jpeg_destroy_decompress(@cinfo);

    { All done. }
    if jerr.num_warnings <> 0 then
    begin
      OutStream.Free;
      raise EJPEGException.Create('There are warnings');
    end
    else
    begin
      OutStream.Position := 0;
      (Picture.Bitmap as TGraphic).LoadFromStream(outstream);
      {I do not know why, but the next line avoids colors to be corrupted on
       256 colors mode}
      Picture.Bitmap.Palette := Picture.Bitmap.Palette;
      OutStream.Free;
    end;
  finally
    { Close files, if we opened them }
    system.close(input_file);
  end;
end;

end.
